#include <avr/io.h>                 
#include <util/delay.h>
#include <avr/interrupt.h>
#include <time.h>
#include <stdlib.h>

#include "usart-lib\usart.h"
#include "avr-nokia5110\nokia5110.h"
#include "spi-lib\spi.h"
#include "lis3dh.h"

#define LOOP_DELAY_MS 10
#define LOOP_DELAY_S 0.01
#define ACC_SENS 1200
#define WALL_MAX_LEN 15
#define WALL_MIN_LEN 2

#define LCD_X 84
#define LCD_Y 48

typedef struct Ball {
	double x;
	double y;
	int r;
	double vx;
	double vy;
} Ball;

typedef struct Wall {
	int x;
	int y;
	int len;
	int dir;
	int exists;
} Wall;

Wall wall = {0, 0, 0, 0, 0};
Ball ball = {0, 0, 2, 0, 0};
Sample sample = {0, 0, 0};

// const for acceleration conversion from accelerometer
const double sens = 0.061f / 1000;

int double_to_int(double value) {
	return (int)(value + 0.5 - (value<0));
}

// end game routine
void lose_message(int score) {
	char score_string[10];
	sprintf(score_string, "%d", score);
	
	nokia_lcd_clear();
	nokia_lcd_write_string("YOU LOSE!", 1);
	nokia_lcd_set_cursor(0, 10);
	nokia_lcd_write_string("SCORE:", 1);
	nokia_lcd_set_cursor(0, 20);
	nokia_lcd_write_string(score_string, 3);
	nokia_lcd_render();
	while(1) {_delay_ms(1000);};
}

// self-explanatory
void check_wall_collision(int score) {
	if (!wall.exists)
		return;
	
	int ballx = double_to_int(ball.x);
	int bally = double_to_int(ball.y);
	int r = ball.r;
	if (wall.dir == 1) {
		int startx = wall.x - wall.len;
		int endx = wall.x + wall.len + 1;
		int wally = wall.y;
		for(int tx = startx; tx<endx; tx++) {
			if((tx >= ballx - r && tx <= ballx + r) && (wally >= bally - r && wally <= bally + r))
				lose_message(score);
		}
	} else if (wall.dir == -1) {
		int starty = wall.y - wall.len;
		int endy = wall.y + wall.len + 1;
		int wallx = wall.x;
		for(int ty = starty; ty<endy; ty++) {
			if((ty >= bally - r && ty <= bally + r) && (wallx >= ballx - r && wallx <= ballx + r))
				lose_message(score);
		}
	}
}

// used for setting or unsetting ball pixels bytes from screen buffer
void draw_ball(int on) {
	int ballx = double_to_int(ball.x);
	int bally = double_to_int(ball.y);
	
	for(int x=ballx-ball.r; x<=ballx+ball.r; x++)
	{
		if (x < 0 || x >= LCD_X)
			continue;
		
		for(int y=bally-ball.r; y<=bally+ball.r; y++)
		{
			if (y < 0 || y >= LCD_Y)
				continue;
			
			nokia_lcd_set_pixel(x, y, on);
		}
	}
}

// used for setting or unsetting wall pixels bytes from screen buffer
void draw_wall(int on) {
	if(!wall.exists)
		return;
	
	int x = wall.x;
	int y = wall.y;
	if (wall.dir == 1) {
		if (y >= 0 && y < LCD_Y)
			for(int i=0; i<= wall.len; i++) {
				int tx1 = x + i;
				int tx2 = x - i;
				if (tx1 >= 0 && tx1 < LCD_X)
					nokia_lcd_set_pixel(tx1, y, on);
				if (tx2 >= 0 && tx2 < LCD_X)
					nokia_lcd_set_pixel(tx2, y, on);
			}
	} else if (wall.dir == -1){
		if (x >= 0 && x < LCD_X)
			for(int i=0; i<= wall.len; i++) {
				int ty1 = y + i;
				int ty2 = y - i;
				if (ty1 >= 0 && ty1 < LCD_Y)
					nokia_lcd_set_pixel(x, ty1, on);
				if (ty2 >= 0 && ty2 < LCD_Y)
					nokia_lcd_set_pixel(x, ty2, on);
			}
	}
}

void move_ball(double newX, double newY) {
	draw_ball(0);
	ball.x = newX;
	ball.y = newY;
	draw_ball(1);
}

/*

  VARIOUS DEBUG TESTS

void lcd_test(void) {
	nokia_lcd_init();
	nokia_lcd_clear();
	
	while(1) {
		nokia_lcd_write_string("IT'S WORKING!",1);
		nokia_lcd_set_cursor(0, 10);
		nokia_lcd_write_string("Nice!", 3);
		nokia_lcd_render();
		_delay_ms(500);
		nokia_lcd_clear();
	};
}

void lcd_test2(void) {
	nokia_lcd_init();
	nokia_lcd_clear();
	
	ball = (Ball){10, 10, 2, 0, 0};
	
	draw_ball(1);
	nokia_lcd_render();
	_delay_ms(10);
	
	
	nokia_lcd_write_string("IT'S WORKING!",1);
	nokia_lcd_set_cursor(0, 10);
	nokia_lcd_write_string("Nice!", 3);
	nokia_lcd_render();
	
	int acc = 10;
	
	while(1) {
		if (acc >= 500) {
			acc = 0;
			move_ball(ball.x + 1, ball.y);
			nokia_lcd_render();
		} else {
			acc += 10;
		}
		_delay_ms(10);
	}
}

void usart_test(void) {
	
	USART0_init();
	
	while(1) {
		USART0_print("HELLO WORLD\n\r");
		_delay_ms(100);
	}
}
 
void spi_test(void) {
	SPI_init();
	USART0_init();
	while(1) {
		uint8_t result = 11;//SPI_read_dummy();
		char buf[64];
		sprintf(buf, "result is: %d\n\r", result);
		USART0_print(buf);
		_delay_ms(1000);
	}
}

void accel_test(void) {
	USART0_init();
	nokia_lcd_init();
	nokia_lcd_clear();
	bool init = lis3dh_init();
	_delay_ms(1);
	char buf[64];
	//char buf2[64];
	sprintf(buf, "init was: %d\n\r", init);
	USART0_print(buf);
	_delay_ms(1000);
	
	
	
	while(1) {
		bool newdata = lis3dh_getSample(&sample);
		if (!newdata) {
			USART0_print("no new data\n\r");
			_delay_ms(500);
		} else {
			char buf3[128];
			sprintf(buf3, "data:: x: %.4f -- y: %.4f -- z: %.4f\n\r", sens * sample.x, sens * sample.y, sens * sample.z);
			USART0_print(buf3);
			_delay_ms(500);
		}
	}
}
*/

void game_loop(void) {
	// init slave devices
	nokia_lcd_init();
	nokia_lcd_clear();
	lis3dh_init();
	srand(time(NULL));
	//srand(0);
	_delay_ms(1);
	
	// startup display
	{
		nokia_lcd_write_string("GET READY!", 1);
		nokia_lcd_set_cursor(43, 21);
		nokia_lcd_write_string("3", 3);
		nokia_lcd_render();
		_delay_ms(1000);
		nokia_lcd_clear();
		nokia_lcd_write_string("GET READY!", 1);
		nokia_lcd_set_cursor(43, 21);
		nokia_lcd_write_string("2", 3);
		nokia_lcd_render();
		_delay_ms(1000);
		nokia_lcd_clear();
		nokia_lcd_write_string("GET READY!", 1);
		nokia_lcd_set_cursor(43, 21);
		nokia_lcd_write_string("1", 3);
		nokia_lcd_render();
		_delay_ms(1000);
		nokia_lcd_clear();
	}
	
	ball = (Ball){41, 23, 2, 0, 0};
	
	draw_ball(1);
	nokia_lcd_render();
	_delay_ms(3000);
	
	int rand_accel_timer = 0;
	double rand_x_accel = 0;
	double rand_y_accel = 0;
	int wall_changed = 0;
	
	int game_time_secs = 0;
	int temp_ms = 0;
	
	while(1) {
		// get accel readings
		lis3dh_getSample(&sample);
		
		// make sure accel doesnt go above aprox. +/- 1g
		double xacc = sens * -sample.x;
		double yacc = sens * -sample.y;
		if (xacc < -1)
			xacc = -1;
		if (xacc > 1)
			xacc = 1;
		if (yacc < -1)
			yacc = -1;
		if (yacc > 1)
			yacc = 1;
		
		// add random accel => this is where the game difficulty comes from
		xacc += rand_x_accel;
		yacc += rand_y_accel;
		
		// do maths and get new position for ball
		double newx = ball.x + ball.vx * LOOP_DELAY_S + (xacc * ACC_SENS) * LOOP_DELAY_S * LOOP_DELAY_S;
		double newy = ball.y + ball.vy * LOOP_DELAY_S + (yacc * ACC_SENS) * LOOP_DELAY_S * LOOP_DELAY_S;
		double newvx = ball.vx + (xacc * ACC_SENS) * LOOP_DELAY_S;
		double newvy = ball.vy + (yacc * ACC_SENS) * LOOP_DELAY_S;
		
		ball.vx = newvx;
		ball.vy = newvy;
		
		int oldx = double_to_int(ball.x);
		int oldy = double_to_int(ball.y);
		int newx_int = double_to_int(newx);
		int newy_int = double_to_int(newy);
		move_ball(newx, newy);
		
		// check if ball is out of bounds and wall collision
		if (newx_int < 0 || newx_int >= LCD_X || newy_int < 0 || newy_int >= LCD_Y)
			lose_message(game_time_secs);
		
		check_wall_collision(game_time_secs);
		
		// 5 secs passed => apply new random acceleration and wall
		if (rand_accel_timer > 5000) {
			rand_accel_timer = 0;
			
			// set random accelerations
			int rand_dir_x = (rand() < RAND_MAX/2) ? -1 : 1;
			int rand_dir_y = (rand() < RAND_MAX/2) ? -1 : 1;
			
			int difficulty_reduction = (61 - game_time_secs);
			if (difficulty_reduction < 1)
				difficulty_reduction = 1;
			
			rand_x_accel = ((double)rand() / (double)RAND_MAX) / difficulty_reduction * rand_dir_x;
			rand_y_accel = ((double)rand() / (double)RAND_MAX) / difficulty_reduction * rand_dir_y;
			
			// set random wall position
			int wallx = newx_int + ball.r + (rand() % (LCD_X - 2 * (ball.r + WALL_MAX_LEN)));
			int wally = newy_int + ball.r + (rand() % (LCD_Y - 2 * (ball.r + WALL_MAX_LEN)));
			if (wallx < 0)
				wallx = LCD_X + wallx;
			if (wallx > LCD_X-1)
				wallx = wallx - LCD_X;
			if (wally < 0)
				wally = LCD_Y + wally;
			if (wally > LCD_Y-1)
				wally = wally - LCD_Y;
			double scale = (double)game_time_secs / 60.0f;
			if (scale > 1)
				scale = 1;
			int walllen = WALL_MIN_LEN + double_to_int(scale * (WALL_MAX_LEN - WALL_MIN_LEN));
			draw_wall(0);
			wall = (Wall){wallx, wally, walllen, rand_dir_x, 1};
			draw_wall(1);
			wall_changed = 1;
		}
		
		// increment game time / score
		if (temp_ms >= 1000) {
			temp_ms = 0;
			game_time_secs += 1;
		}
		
		// render screen only if ball or wall moved
		if ((oldx != newx_int) || (oldy != newy_int) || wall_changed) {
			nokia_lcd_render();
			wall_changed = 0;
		}
		
		// loop delay
		_delay_ms(LOOP_DELAY_MS);
		rand_accel_timer += LOOP_DELAY_MS;
		temp_ms += LOOP_DELAY_MS;
	}
}

int main(void) 
{
	game_loop();
}
