// Official project location:
// https://github.com/rickkas7/LIS3DH

// modified by Octav Rusuleanu for PM project

#include <avr/io.h>
#include <util/delay.h>

#include "lis3dh.h"
#include "spi-lib/spi.h"

bool lis3dh_init() {
	SPI_init();
	_delay_ms(1);
	
	if(!lis3dh_hasDevice())
		return false;
	
	// 400 Hz, no low power, x y z enabled
	uint8_t reg1 = 0b01110111;
	
	// ctrl reg for high pass filter settings - will not use it
	//uint8_t reg2 = 0;
	
	// ctrl reg for INT settings - will not use it
	//uint8_t reg3 = 0;
	
	// ctrl reg for various settings - bits 5, 6 used for full scale setup - leave default
	//uint8_t reg4 = 0;
	
	// ctrl registers 5 and 6 outside the scope of the project
	
	lis3dh_writeReg8(REG_CTRL_REG1, reg1);
	
	return true;
}

bool lis3dh_getSample(Sample *sample) {
	uint8_t statusAuxReg = lis3dh_readReg8(REG_STATUS_REG);

	bool hasData = ((statusAuxReg & STATUS_ZYXDA) != 0);

	//Serial.printlnf("fifoSrcReg=0x%02x", fifoSrcReg);

	if (hasData) {
		uint8_t resp[6];
		lis3dh_readData(REG_OUT_X_L, resp, sizeof(resp));

		sample->x = (int16_t) (resp[0] | (((uint16_t)resp[1]) << 8));
		sample->y = (int16_t) (resp[2] | (((uint16_t)resp[3]) << 8));
		sample->z = (int16_t) (resp[4] | (((uint16_t)resp[5]) << 8));
	}
	return hasData;
}

bool lis3dh_hasDevice() {
	bool found = false;
	for(int tries = 0; tries < 10; tries++) {
		uint8_t whoami = lis3dh_readReg8(REG_WHO_AM_I);
		if (whoami == WHO_AM_I) {
			found = true;
			break;
		}
		_delay_ms(1);
	}
	return found;
}

bool lis3dh_readData(uint8_t addr, uint8_t *buf, size_t numBytes) {
	SPI_begin_ss();

	if (numBytes > 1) {
		addr |= _SPI_INCREMENT;
	}

	SPI_exchange(_READ_INSTR | addr);

	for(size_t ii = 0; ii < numBytes; ii++) {
		buf[ii] = SPI_exchange(0);
	}

	SPI_end_ss();

	return true;
}

bool lis3dh_writeData(uint8_t addr, const uint8_t *buf, size_t numBytes) {
	SPI_begin_ss();

	if (numBytes > 1) {
		addr |= _SPI_INCREMENT;
	}

	// _WRITE_INSTR is = 0 so need to mask it
	SPI_exchange(addr);
	for(size_t ii = 0; ii < numBytes; ii++) {
		SPI_exchange(buf[ii]);
	}

	SPI_end_ss();

	return true;
}

uint8_t lis3dh_readReg8(uint8_t addr) {

	uint8_t resp[1];
	lis3dh_readData(addr, resp, sizeof(resp));

	return resp[0];
}

uint16_t lis3dh_readReg16(uint8_t addr) {

	uint8_t resp[2];
	lis3dh_readData(addr, resp, sizeof(resp));

	return resp[0] | (((uint16_t)resp[1]) << 8);
}


void lis3dh_writeReg8(uint8_t addr, uint8_t value) {
	// Serial.printlnf("writeRegister addr=%02x value=%02x", addr, value);

	uint8_t req[1];
	req[0] = value;

	lis3dh_writeData(addr, req, sizeof(req));
}

void lis3dh_writeReg16(uint8_t addr, uint16_t value) {
	// Serial.printlnf("writeRegister addr=%02x value=%04x", addr, value);

	uint8_t req[2];
	req[0] = value & 0xff;
	req[1] = value >> 8;

	lis3dh_writeData(addr, req, sizeof(req));
}